/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

package com.sunlabs.archsim.fleet.ships.tokenFifo;

import com.sunlabs.archsim.base.common.Parameters;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.FleetGlobal;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.State;
import com.sunlabs.archsim.gasp.Terminal;

/**
 * A Fifo with oken interface 
 * 
 * @author ib27688
 *
 */
public class TokenFifoShip extends Component {
	
	protected TokenFifoShipElement first;
	protected TokenFifoShipElement last;
	
	// used by XML interface
	public TokenFifoShip(String name, Parameters params) {
		super(name, null);
		
		// process the parameters
		Delay fwdDelayParam= null;
		Delay ackDelayParam= null;
		Delay tokenDelayParam= null;
		Delay ackTokenDelayParam= null;
		int capacityParam= TokenFifoShipDefaults.TOKEN_FIFO_SHIP_CAPACITY_DEF;
		
		if (params != null) {
			for (int i= 0; i<params.size(); i++) {
				String pName= params.getName(i);
				String sValue= params.getValue(i);
				
				if (pName.equals(TokenFifoShipDefaults.TOKEN_FIFO_SHIP_FWD_DELAY) 
						|| pName.equals(TokenFifoShipDefaults.FWD_DELAY)) {
					try {
						int d = Integer.parseInt(sValue);
						fwdDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ TokenFifoShipDefaults.TOKEN_FIFO_SHIP_FWD_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED");						
					}
				}
				else if (pName.equals(TokenFifoShipDefaults.TOKEN_FIFO_SHIP_ACK_DELAY) 
						|| pName.equals(TokenFifoShipDefaults.ACK_DELAY)) {
					try {
						int d = Integer.parseInt(sValue);
						ackDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ TokenFifoShipDefaults.TOKEN_FIFO_SHIP_ACK_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED");						
					}
				}
				else if (pName.equals(TokenFifoShipDefaults.TOKEN_DELAY)) {
					try {
						int d = Integer.parseInt(sValue);
						tokenDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ TokenFifoShipDefaults.TOKEN_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED");						
					}
				}
				else if (pName.equals(TokenFifoShipDefaults.ACK_TOKEN_DELAY)) {
					try {
						int d = Integer.parseInt(sValue);
						ackTokenDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ TokenFifoShipDefaults.ACK_TOKEN_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED");						
					}
				}
				if (pName.equals(TokenFifoShipDefaults.TOKEN_FIFO_SHIP_CAPACITY) 
						|| pName.equals(FleetGlobal.PARAMETER) ) {
					try {
						capacityParam = Integer.parseInt(sValue);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ TokenFifoShipDefaults.TOKEN_FIFO_SHIP_CAPACITY
								+ ": " + sValue
								+ ". Value is IGNORED");						
					}
				}				
				else {
					director.error("Component: " + name + ", id= " + getID() 
							+ ": Attribute " + pName + " with value " + sValue
							+ " is ignored");					
				} // else
			} // for
		} // if params != null			
		
		buildTokenFifoShip(capacityParam, 
				fwdDelayParam, ackDelayParam, tokenDelayParam, ackTokenDelayParam);
		
	} // TokenFifoShipConstructor
	
	public TokenFifoShip(String name, int capacity) {
		super(name, null);

		buildTokenFifoShip(capacity, null, null, null, null);

	} // TokenFifoConstructor
	
	protected void buildTokenFifoShip(int capacity, 
			Delay fwdDelay, Delay ackDelay, Delay tokenDelay, Delay ackTokenDelay) {
		// no negative capacity - make a zero capacity fifo
		if (capacity < 0) capacity= 0;
		
		// make the first element 
		first= new TokenFifoShipElement(0);
		last= first; // need this for capacity 0 
		addMember(first);
		
		for (int i=0; i< capacity; i++) {
			// make the next F
			TokenFifoShipElement next=  new TokenFifoShipElement(i+1);
			addMember(next);
			
			// set delays if needed
			if (fwdDelay != null) {
				next.action.passDataFn.setDelay(fwdDelay);
			}
			if (ackDelay != null) {
				next.action.ackInFn.setDelay(ackDelay);
			}				
			if (tokenDelay != null) {
				next.action.tokenOutFn.setDelay(fwdDelay);
			}				
			if (ackTokenDelay != null) {
				next.action.ackTokenInFn.setDelay(fwdDelay);
			}				
			
			// make states and connect them to terminals
			State dataState= new State(FleetGlobal.DATA_STATE);
			addMember(dataState);
			dataState.attach(last.out);
			dataState.attach(next.in);
			
			// ackState must b initially active
			State ackState= new State(FleetGlobal.ACK_STATE);
			addMember(ackState);
			ackState.attach(last.ackOut);
			ackState.attach(next.ackIn);
			// ackState must be active at initialization/reset
			ackState.setInitiallyActive(new Data());
			
			// now connect tokens
			State tokenState= new State(FleetGlobal.TOKEN_STATE);
			addMember(tokenState);
			tokenState.attach(last.tokenIn);
			tokenState.attach(next.tokenOut);
			// token state is initially active
			tokenState.setInitiallyActive(new Data());

			State ackTokenState= new State(FleetGlobal.ACK_TOKEN_STATE);
			addMember(ackTokenState);
			ackTokenState.attach(last.ackTokenIn);
			ackTokenState.attach(next.ackTokenOut);
				
			// prepare for the next iteration
			last= next;
		}		
	}
	
	public Terminal getTerminal(String name) {
		if (name.equals(FleetGlobal.IN)) {
			return first.in;
		}
		else if (name.equals(FleetGlobal.ACK_IN)) {
			return first.ackIn;
		}
		else if (name.equals(FleetGlobal.OUT)) {
			return last.out;
		}
		else if (name.equals(FleetGlobal.ACK_OUT)) {
			return last.ackOut;
		}
		if (name.equals(FleetGlobal.TOKEN_IN)) {
			return last.tokenIn;
		}
		else if (name.equals(FleetGlobal.ACK_TOKEN_IN)) {
			return last.ackTokenIn;
		}
		else if (name.equals(FleetGlobal.TOKEN_OUT)) {
			return first.tokenOut;
		}
		else if (name.equals(FleetGlobal.ACK_TOKEN_OUT)) {
			return first.ackTokenOut;
		}
		else
			return null;		
	} // getTerminal

} // TokenFifoShip
