// Copyright 2006 all rights reserved; see LICENSE file for BSD-style license

package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.Sequence.Position;
import java.io.*;
import java.util.*;

/** a parser which translates an Input&lt;Token&gt; into a Forest&lt;NodeType&gt; */
public abstract class Parser<Token, NodeType> {
    protected final Table<Token> pt;

    /** create a parser to parse the grammar with start symbol <tt>u</tt> */
    public Parser(Union u, Topology<Token> top)  { this.pt = new Table<Token>(u, top); }
    Parser(Table<Token> pt)               { this.pt = pt; }

    /** implement this method to create the output forest corresponding to a lone shifted input token */
    public abstract Forest<NodeType> shiftToken(Token t, Input.Location newloc);

    boolean helpgc = true;

    public String toString() { return pt.toString(); }

    /** parse <tt>input</tt>, and return the shared packed parse forest (or throw an exception) */
    public Forest<NodeType> parse(Input<Token> input) throws IOException, ParseFailed {
        GSS gss = new GSS(input);
        Input.Location loc = input.getLocation();
        Token tok = input.next();
        GSS.Phase current = gss.new Phase<Token>(null, null, tok, loc, input.getLocation(), null);
        current.newNode(new Result(Forest.create(loc.createRegion(loc), null, null, false), null, null), pt.start, true);
        int count = 1;
        for(int idx=0;;idx++) {
            Input.Location oldloc = loc;
            current.reduce();
            Forest forest = current.token==null ? null : shiftToken((Token)current.token, loc);
            loc = input.getLocation();
            Token nextToken = input.next();
            GSS.Phase next = gss.new Phase<Token>(current, current, nextToken, loc, input.getLocation(), forest);

            /*
            FileOutputStream fos = new FileOutputStream("out-"+idx+".dot");
            PrintWriter p = new PrintWriter(new OutputStreamWriter(fos));
            GraphViz gv = new GraphViz();
            for(Object n : current)
                ((Node)n).toGraphViz(gv);
            gv.dump(p);
            p.flush();
            p.close();
            */

            count = next.size();
            if (current.isDone()) return (Forest<NodeType>)gss.finalResult;
            current = next;
        }
    }

    // Table //////////////////////////////////////////////////////////////////////////////

    /** an SLR(1) parse table which may contain conflicts */
    static class Table<Token> extends Walk.Cache {

        public String toString() {
            StringBuffer sb = new StringBuffer();
            sb.append("parse table");
            for(State<Token> state : all_states) {
                sb.append("  " + state + "\n");
                for(Topology<Token> t : state.shifts) {
                    sb.append("      shift  \""+
                              new edu.berkeley.sbp.chr.CharTopology((IntegerTopology<Character>)t)+"\" => ");
                    for(State st : state.shifts.getAll(t))
                        sb.append(st.idx+"  ");
                    sb.append("\n");
                }
                for(Topology<Token> t : state.reductions)
                    sb.append("      reduce \""+
                              new edu.berkeley.sbp.chr.CharTopology((IntegerTopology<Character>)t)+"\" => " +
                              state.reductions.getAll(t) + "\n");
                for(Sequence s : state.gotoSetNonTerminals.keySet())
                    sb.append("      goto   "+state.gotoSetNonTerminals.get(s)+" from " + s + "\n");
            }
            return sb.toString();
        }

        public final Walk.Cache cache = this;

        private void walk(Element e, HashSet<SequenceOrElement> hs) {
            if (e==null) return;
            if (hs.contains(e)) return;
            hs.add(e);
            if (e instanceof Atom) return;
            for(Sequence s : (Union)e)
                walk(s, hs);
        }
        private void walk(Sequence s, HashSet<SequenceOrElement> hs) {
            hs.add(s);
            for(Position p = s.firstp(); p != null; p = p.next())
                walk(p.element(), hs);
            for(Sequence ss : s.needs()) walk(ss, hs);
            for(Sequence ss : s.hates()) walk(ss, hs);
        }

        /** the start state */
        public  final State<Token>   start;

        /** the state from which no reductions can be done */
        private final State<Token>   dead_state;

        /** used to generate unique values for State.idx */
        private int master_state_idx = 0;
        HashSet<State<Token>>   all_states    = new HashSet<State<Token>>();
        HashMap<HashSet<Position>,State<Token>>   doomed_states    = new HashMap<HashSet<Position>,State<Token>>();
        HashMap<HashSet<Position>,State<Token>>   normal_states    = new HashMap<HashSet<Position>,State<Token>>();
        HashSet<SequenceOrElement>                all_elements  = new HashSet<SequenceOrElement>();

        /** construct a parse table for the given grammar */
        public Table(Topology top) { this("s", top); }
        public Table(String startSymbol, Topology top) { this(new Union(startSymbol), top); }
        public Table(Union ux, Topology top) {
            Union start0 = new Union("0");
            start0.add(new Sequence.Singleton(ux));

            for(Sequence s : start0) cache.eof.put(s, true);
            cache.eof.put(start0, true);

            // construct the set of states
            walk(start0, all_elements);
            for(SequenceOrElement e : all_elements)
                cache.ys.addAll(e, new Walk.YieldSet(e, cache).walk());
            for(SequenceOrElement e : all_elements)
                cache.ys2.addAll(e, new Walk.YieldSet2(e, cache).walk());
            HashSet<Position> hp = new HashSet<Position>();
            reachable(start0, hp);

            this.dead_state = new State<Token>(new HashSet<Position>(), true);
            this.start = new State<Token>(hp);

            // for each state, fill in the corresponding "row" of the parse table
            for(State<Token> state : all_states)
                for(Position p : state.hs) {

                    // the Grammar's designated "last position" is the only accepting state
                    if (start0.contains(p.owner()) && p.next()==null && !state.doomed)
                        state.accept = true;

                    if (isRightNullable(p)) {
                        Walk.Follow wf = new Walk.Follow(top.empty(), p.owner(), all_elements, cache);
                        Topology follow = wf.walk(p.owner());
                        for(Position p2 = p; p2 != null && p2.element() != null; p2 = p2.next()) {
                            Atom set = new Walk.EpsilonFollowSet(new edu.berkeley.sbp.chr.CharAtom(top.empty().complement()),
                                                                 new edu.berkeley.sbp.chr.CharAtom(top.empty()),
                                                                 cache).walk(p2.element());
                            follow = follow.intersect(new Walk.Follow(top.empty(), p2.element(), all_elements, cache).walk(p2.element()));
                            if (set != null) follow = follow.intersect(set.getTokenTopology());
                        }
                        state.reductions.put(follow, p);
                        if (wf.includesEof()) state.eofReductions.add(p);
                    }

                    // if the element following this position is an atom, copy the corresponding
                    // set of rows out of the "master" goto table and into this state's shift table
                    if (p.element() != null && p.element() instanceof Atom)
                        state.shifts.addAll(state.gotoSetTerminals.subset(((Atom)p.element()).getTokenTopology()));
                }
            if (top instanceof IntegerTopology)
                for(State<Token> state : all_states) {
                    state.oreductions = state.reductions.optimize(((IntegerTopology)top).functor());
                    state.oshifts = state.shifts.optimize(((IntegerTopology)top).functor());
                }
        }

        private boolean isRightNullable(Position p) {
            if (p.isLast()) return true;
            if (!possiblyEpsilon(p.element())) return false;
            return isRightNullable(p.next());
        }

        /** a single state in the LR table and the transitions possible from it */

        class State<Token> implements IntegerMappable, Iterable<Position> {
        
            public  final     int               idx    = master_state_idx++;
            private final     HashSet<Position> hs;
            public HashSet<State<Token>> also = new HashSet<State<Token>>();

            public transient HashMap<Sequence,State<Token>>         gotoSetNonTerminals = new HashMap<Sequence,State<Token>>();
            private transient TopologicalBag<Token,State<Token>>     gotoSetTerminals    = new TopologicalBag<Token,State<Token>>();

            private           TopologicalBag<Token,Position> reductions          = new TopologicalBag<Token,Position>();
            private           HashSet<Position>              eofReductions       = new HashSet<Position>();
            private           TopologicalBag<Token,State<Token>>     shifts              = new TopologicalBag<Token,State<Token>>();
            private           boolean                         accept              = false;

            private VisitableMap<Token,State<Token>> oshifts = null;
            private VisitableMap<Token,Position> oreductions = null;

            // Interface Methods //////////////////////////////////////////////////////////////////////////////

            boolean             isAccepting()           { return accept; }
            public Iterator<Position>  iterator()       { return hs.iterator(); }

            boolean             canShift(Token t)         { return oshifts!=null && oshifts.contains(t); }
            <B,C> void          invokeShifts(Token t, Invokable<State<Token>,B,C> irbc, B b, C c) {
                oshifts.invoke(t, irbc, b, c);
            }

            boolean             canReduce(Token t)        { return oreductions != null && (t==null ? eofReductions.size()>0 : oreductions.contains(t)); }
            <B,C> void          invokeReductions(Token t, Invokable<Position,B,C> irbc, B b, C c) {
                if (t==null) for(Position r : eofReductions) irbc.invoke(r, b, c);
                else         oreductions.invoke(t, irbc, b, c);
            }

            // Constructor //////////////////////////////////////////////////////////////////////////////

            /**
             *  create a new state consisting of all the <tt>Position</tt>s in <tt>hs</tt>
             *  @param hs           the set of <tt>Position</tt>s comprising this <tt>State</tt>
             *  @param all_elements the set of all elements (Atom instances need not be included)
             *  
             *   In principle these two steps could be merged, but they
             *   are written separately to highlight these two facts:
             * <ul>
             * <li> Non-atom elements either match all-or-nothing, and do not overlap
             *      with each other (at least not in the sense of which element corresponds
             *      to the last reduction performed).  Therefore, in order to make sure we
             *      wind up with the smallest number of states and shifts, we wait until
             *      we've figured out all the token-to-position multimappings before creating
             *      any new states
             *  
             * <li> In order to be able to run the state-construction algorithm in a single
             *      shot (rather than repeating until no new items appear in any state set),
             *      we need to use the "yields" semantics rather than the "produces" semantics
             *      for non-Atom Elements.
             *  </ul>
             */
            public State(HashSet<Position> hs) { this(hs, false); }
            public boolean doomed;
            public State(HashSet<Position> hs, boolean doomed) {
                this.hs = hs;
                this.doomed = doomed;

                // register ourselves in the all_states hash so that no
                // two states are ever created with an identical position set
                ((HashMap)(doomed ? doomed_states : normal_states)).put(hs, this);
                ((HashSet)all_states).add(this);
                
                for(Position p : hs) {
                    if (!p.isFirst()) continue;
                    for(Sequence s : p.owner().needs()) {
                        if (hs.contains(s.firstp())) continue;
                        HashSet<Position> h2 = new HashSet<Position>();
                        reachable(s.firstp(), h2);
                        also.add(mkstate(h2, true));
                    }
                    for(Sequence s : p.owner().hates()) {
                        if (hs.contains(s.firstp())) continue;
                        HashSet<Position> h2 = new HashSet<Position>();
                        reachable(s, h2);
                        also.add(mkstate(h2, true));
                    }
                }

                // Step 1a: examine all Position's in this state and compute the mappings from
                //          sets of follow tokens (tokens which could follow this position) to sets
                //          of _new_ positions (positions after shifting).  These mappings are
                //          collectively known as the _closure_

                TopologicalBag<Token,Position> bag0 = new TopologicalBag<Token,Position>();
                for(Position position : hs) {
                    if (position.isLast() || !(position.element() instanceof Atom)) continue;
                    Atom a = (Atom)position.element();
                    HashSet<Position> hp = new HashSet<Position>();
                    reachable(position.next(), hp);
                    bag0.addAll(a.getTokenTopology(), hp);
                }

                // Step 1b: for each _minimal, contiguous_ set of characters having an identical next-position
                //          set, add that character set to the goto table (with the State corresponding to the
                //          computed next-position set).

                for(Topology<Token> r : bag0) {
                    HashSet<Position> h = new HashSet<Position>();
                    for(Position p : bag0.getAll(r)) h.add(p);
                    ((TopologicalBag)gotoSetTerminals).put(r, mkstate(h, doomed));
                }

                // Step 2: for every non-Atom element (ie every Element which has a corresponding reduction),
                //         compute the closure over every position in this set which is followed by a symbol
                //         which could yield the Element in question.
                //
                //         "yields" [in one or more step] is used instead of "produces" [in exactly one step]
                //         to avoid having to iteratively construct our set of States as shown in most
                //         expositions of the algorithm (ie "keep doing XYZ until things stop changing").

                HashMapBag<SequenceOrElement,Position> move = new HashMapBag<SequenceOrElement,Position>();
                for(Position p : hs) {
                    Element e = p.element();
                    if (e==null) continue;
                    for(SequenceOrElement y : cache.ys2.getAll(e)) {
                        //System.out.println(e + " yields " + y);
                        HashSet<Position> hp = new HashSet<Position>();
                        reachable(p.next(), hp);
                        move.addAll(y, hp);
                    }
                }
                OUTER: for(SequenceOrElement y : move) {
                    HashSet<Position> h = move.getAll(y);
                    State<Token> s = mkstate(h, doomed);
                    // if a reduction is "lame", it should wind up in the dead_state after reducing
                    if (y instanceof Sequence) {
                        for(Position p : hs) {
                            if (p.element() != null && (p.element() instanceof Union)) {
                                Union u = (Union)p.element();
                                for(Sequence seq : u)
                                    if (seq.needs.contains((Sequence)y) || seq.hates.contains((Sequence)y)) {
                                        // FIXME: what if there are two "routes" to get to the sequence?
                                        ((HashMap)gotoSetNonTerminals).put((Sequence)y, dead_state);
                                        continue OUTER;
                                    }
                            }
                        }
                        gotoSetNonTerminals.put((Sequence)y, s);
                    }
                }
            }

            private State<Token> mkstate(HashSet<Position> h, boolean b) {
                if (b) return doomed_states.get(h) == null ? (State)new State<Token>(h,b) : (State)doomed_states.get(h);
                else   return normal_states.get(h) == null ? (State)new State<Token>(h,b) : (State)normal_states.get(h);
            }

            public String toStringx() {
                StringBuffer st = new StringBuffer();
                for(Position p : this) {
                    if (st.length() > 0) st.append("\n");
                    st.append(p);
                }
                return st.toString();
            }
            public String toString() {
                StringBuffer ret = new StringBuffer();
                ret.append("state["+idx+"]: ");
                for(Position p : this) ret.append("{"+p+"}  ");
                return ret.toString();
            }

            public Walk.Cache cache() { return cache; }
            public int toInt() { return idx; }
        }
    }

    // Helpers //////////////////////////////////////////////////////////////////////////////
    
    private static void reachable(Sequence s, HashSet<Position> h) {
        reachable(s.firstp(), h);
        //for(Sequence ss : s.needs()) reachable(ss, h);
        //for(Sequence ss : s.hates()) reachable(ss, h);
    }
    private static void reachable(Element e, HashSet<Position> h) {
        if (e instanceof Atom) return;
        for(Sequence s : ((Union)e))
            reachable(s, h);
    }
    private static void reachable(Position p, HashSet<Position> h) {
        if (h.contains(p)) return;
        h.add(p);
        if (p.element() != null) reachable(p.element(), h);
    }

}
